/*
   Copyright (c) JRL CNRS-AIST,
   @author :
                Olivier STASSE
                Alireza NAKHAEI

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#include "worldModelGrid3D/wmGrid3D.h"

//-------------------------------------------------------------------------
// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "wmGrid3D : " << x << std::endl
#define ODEBUG1(x) std::cerr << "wmGrid3D:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "wmGrid3D:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

#define MINIMUM_ACCEPTABLE_OBSTACLE_DISTANCE 0.6
#define CAMERA_HEIGHT 1.29
#define ROBOT_X 0.6
#define ROBOT_Y 0.9

using namespace std;

// constructor
CwmGrid3D::CwmGrid3D()
{
#ifdef WM3D_HAS_BOOST
	attSensorPosition.resize ( 4,4 );
#endif
	attGridMap = NULL;
	attResolution = DEFAULT_RESOLUTION;
	attSensitivity = DEFAULT_SENSITIVITY;

	if ( allocateMap ( DEFAULT_MAP_AREA_XSTART, DEFAULT_MAP_AREA_YSTART,
	                   DEFAULT_MAP_AREA_ZSTART, DEFAULT_MAP_AREA_XEND,
	                   DEFAULT_MAP_AREA_YEND, DEFAULT_MAP_AREA_ZEND ) !=WM_OK )
	{
		ODEBUG1 ( "ERROR : CwmGrid3D::CwmGrid3D() - error in allocateMap " );

	}
}

CwmGrid3D::~CwmGrid3D()
{
	if ( attGridMap )
		delete [] attGridMap;

}

inline grid *CwmGrid3D::xyz2grid ( double x, double y, double z )
{

	int xx= ( int ) rint ( ( x-attMapArea.xstart ) /attResolution );
	int yy= ( int ) rint ( ( y-attMapArea.ystart ) /attResolution );
	int zz= ( int ) rint ( ( z-attMapArea.zstart ) /attResolution );

	//int xx=floor ( ( x-attMapArea.xstart ) /attResolution+0.001 );
	//int yy=floor ( ( y-attMapArea.ystart ) /attResolution +0.001);
	//int zz=floor ( ( z-attMapArea.zstart ) /attResolution+0.001 );

	if ( xx==attMapXN )
		xx=xx-1;
	if ( yy==attMapYN )
		yy=yy-1;
	if ( zz==attMapZN )
		zz=zz-1;
	return xyz2grid ( xx,yy,zz );
}


inline grid *CwmGrid3D::xyz2grid ( int x, int y, int z )
{

	return attGridMap + x + y*attMapXN + z*attMapXN*attMapYN;
}

wmStatus CwmGrid3D::updateMap ( )
{
	grid *g = attGridMap;
	g = xyz2grid ( attMapArea.xstart, attMapArea.ystart, attMapArea.zstart );
	double step=attResolution/2;

#ifdef WM3D_HAS_BOOST  //using boost
	ODEBUG2 ( "USING LIB-BOOST " );
	boost::numeric::ublas::vector<float> PosInGlobal ( 4 ),PosInStereo ( 4 );
	int lOccupied =0;
	int inNbOfPixel=attRangeMap.size1();
	ODEBUG2 ( "Number of pixel from Boost Rangemap = " << inNbOfPixel );

	for ( int i=0 ; i<inNbOfPixel;i++ )
	{
		PosInStereo ( 0 ) = attRangeMap ( i,0 );
		PosInStereo ( 1 ) = attRangeMap ( i,1 );
		PosInStereo ( 2 ) = attRangeMap ( i,2 );
		PosInStereo ( 3 ) = 1.0;

		if ( ( PosInStereo ( 0 ) !=-1.0 ) && ( PosInStereo ( 1 ) !=-1.0 ) && ( PosInStereo ( 2 ) !=-1.0 ) )
		{
			PosInGlobal=boost::numeric::ublas::prod ( attSensorPosition,PosInStereo );
			if ( PosInGlobal ( 0 ) >attMapArea.xstart && PosInGlobal ( 0 ) <attMapArea.xend && PosInGlobal ( 1 ) >attMapArea.ystart && PosInGlobal ( 1 ) < attMapArea.yend && PosInGlobal ( 2 ) > attMapArea.zstart && PosInGlobal ( 2 ) < attMapArea.zend )
			{
				double x=PosInGlobal ( 0 );
				double y=PosInGlobal ( 1 );
				double z=PosInGlobal ( 2 );
				g = xyz2grid ( x,y,z );
				if ( g->occupied < 126 )
				{
					g->occupied++;
				}
				g->variance=attRangeMap ( i,3 );
				lOccupied++;
			}
		}
	}

	double angleStep=0.5*atan ( attResolution/attObservableDistance );
	int hideFlag = 0;

	// just for inithializing the parameters for using in the while
	double xx= ( attMapArea.xstart+attMapArea.xend ) /2;
	double yy= ( attMapArea.ystart+attMapArea.yend ) /2;
	double zz= ( attMapArea.zstart+attMapArea.zend ) /2;

	grid *gg = attGridMap;

	for ( double j=-1*attHeightAngle/2 ; j<attHeightAngle/2; j=j+angleStep )
	{
		for ( double i=-1*attWidthAngle/2 ; i<attWidthAngle/2; i=i+angleStep )
		{
			double inMean =1000;
			double inVariance=0;
			double intenseFactor = 0;
			double rr=0.02;
			while ( rr < attObservableDistance && hideFlag ==0 )
			{
				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );
				PosInStereo ( 3 ) =1;
				PosInGlobal=boost::numeric::ublas::prod ( attSensorPosition,PosInStereo );
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{
					gg = xyz2grid ( xx, yy , zz );
					if ( gg->occupied > attSensitivity )
					{
						hideFlag=1;
						inMean = rr;
						inVariance=gg->variance;
						intenseFactor = gg->occupied;
					}
				}
				rr=rr+step;
			}
			hideFlag=0;
//-----------------------------------------------
			rr=0.02;
			while ( rr < attObservableDistance && rr < ( inMean +2*step ) )
			{
				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );
				PosInStereo ( 3 ) =1;
				PosInGlobal=boost::numeric::ublas::prod ( attSensorPosition,PosInStereo );
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{
					gg = xyz2grid ( xx, yy , zz );
					double observedOdd = getGridParameter ( inMean,inVariance,rr, intenseFactor-attSensitivity );
					double currentOdd= gg->odd;
					double newOdd= currentOdd * observedOdd;
					if ( newOdd<10e-6 )
						newOdd=10e-6;
					if ( newOdd>10e6 )
						newOdd=10e6;
					gg->odd= newOdd;
				}
				rr=rr+step;
			}

			xx= ( attMapArea.xstart+attMapArea.xend ) /2;
			yy= ( attMapArea.ystart+attMapArea.yend ) /2;
			zz= ( attMapArea.zstart+attMapArea.zend ) /2;
		}
	}

	// This Time, just for setting all the accupied to zero
	xx= ( attMapArea.xstart+attMapArea.xend ) /2;
	yy= ( attMapArea.ystart+attMapArea.yend ) /2;
	zz= ( attMapArea.zstart+attMapArea.zend ) /2;

	for ( double j=-1*attHeightAngle/2 ; j<attHeightAngle/2; j=j+angleStep )
	{
		for ( double i=-1*attWidthAngle/2 ; i<attWidthAngle/2; i=i+angleStep )
		{
			for ( double rr=0.02 ; rr < attObservableDistance; rr=rr+step )
			{

				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );
				PosInStereo ( 3 ) =1;
				PosInGlobal=boost::numeric::ublas::prod ( attSensorPosition,PosInStereo );
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{
					gg = xyz2grid ( xx, yy , zz );
					gg->occupied = 0;
				}
			}
			xx= ( attMapArea.xstart+attMapArea.xend ) /2;
			yy= ( attMapArea.ystart+attMapArea.yend ) /2;
			zz= ( attMapArea.zstart+attMapArea.zend ) /2;
		}
	}

#endif  //using boost

#ifdef WM3D_HAS_LIBIMAGES3D   //using libimage3D
	ODEBUG2 ( "USING LIB-IMAGE-3D " );
	V3D PosInGlobal;
	V3D PosInStereo;
	int lOccupied =0;
	int lOccupied_total =0;

	int nbLine = attRangeMap->header.nblig;
	int nbCol = attRangeMap->header.nbcol;

	int inNbOfPixel=nbLine*nbCol;
	ODEBUG2 ( "nbLine = " << nbLine );
	ODEBUG2 ( "nbCol = " << nbCol );
	ODEBUG2 ( "Number of pixel from lib3D = " << inNbOfPixel );

	ODEBUG2 ( "SENSOR POS BEFORE = " << attSensorPosition );
	for ( int i=0 ; i<inNbOfPixel;i++ )
	{
		if ( attRangeMap->points_tab[i].state == GOOD_PT3D )
		{
			PosInStereo ( 0 ) = attRangeMap->points_tab[i].coord_1;
			PosInStereo ( 1 ) = attRangeMap->points_tab[i].coord_2;
			PosInStereo ( 2 ) = attRangeMap->points_tab[i].coord_3;


			PosInGlobal=attSensorPosition*PosInStereo;
			if ( PosInGlobal ( 0 ) >attMapArea.xstart && PosInGlobal ( 0 ) <attMapArea.xend && PosInGlobal ( 1 ) >attMapArea.ystart && PosInGlobal ( 1 ) < attMapArea.yend && PosInGlobal ( 2 ) > attMapArea.zstart && PosInGlobal ( 2 ) < attMapArea.zend )
			{
				double x=PosInGlobal ( 0 );
				double y=PosInGlobal ( 1 );
				double z=PosInGlobal ( 2 );
				g = xyz2grid ( x,y,z );
				if ( g->occupied < 126 )
				{
					g->occupied++;
				}
				g->variance= ( attRangeMap->points_tab[i].covariances_tab[5] );
				lOccupied++;

			}
			lOccupied_total++;
		}
	}
	ODEBUG2 ( "Total number of occupied grid = " << lOccupied_total );
	ODEBUG2 ( "number of occupied grid in the image= " << lOccupied );

	std::vector<double> inGridParameters;
	std::vector<double> currentGridParameters;
	std::vector<double> newGridParameters;

	inGridParameters.resize ( 3 );
	currentGridParameters.resize ( 3 );
	newGridParameters.resize ( 3 );



	//double angleStep=PI/180;
	double angleStep=0.5*atan ( attResolution/attObservableDistance );
	int hideFlag = 0;

	// just for inithializing the parameters for using in the while
	double xx= ( attMapArea.xstart+attMapArea.xend ) /2;
	double yy= ( attMapArea.ystart+attMapArea.yend ) /2;
	double zz= ( attMapArea.zstart+attMapArea.zend ) /2;

	grid *gg = attGridMap;
	for ( double j=-1*attHeightAngle/2 ; j<attHeightAngle/2; j=j+angleStep )
	{

		for ( double i=-1*attWidthAngle/2 ; i<attWidthAngle/2; i=i+angleStep )
		{
			double inMean =1000;
			double inVariance=0;
			double intenseFactor = 0;
			double rr=0.02;

			while ( rr < attObservableDistance && hideFlag ==0 )
			{
				/*				PosInStereo ( 0 ) =rr*cos ( j ) *cos ( i );
								PosInStereo ( 1 ) =rr*cos ( j ) *sin ( i );
								PosInStereo ( 2 ) =rr*sin ( j );*/
				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );
				PosInStereo ( 3 ) =1;
				PosInGlobal=attSensorPosition*PosInStereo;
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{
					gg = xyz2grid ( xx, yy , zz );
					if ( gg->occupied > attSensitivity )
					{

						hideFlag=1;
						inMean = rr;
						inVariance=gg->variance;
						intenseFactor = gg->occupied;
					}
				}
				rr=rr+step;
			}
			hideFlag=0;

//-----------------------------------------------

			rr=0.02;

			while ( rr < attObservableDistance && rr < ( inMean +3*step ) )
			{
				/*				PosInStereo ( 0 ) =rr*cos ( j ) *cos ( i );
								PosInStereo ( 1 ) =rr*cos ( j ) *sin ( i );
								PosInStereo ( 2 ) =rr*sin ( j );*/
				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );

				PosInStereo ( 3 ) =1;
				PosInGlobal=attSensorPosition*PosInStereo ;
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{

					gg = xyz2grid ( xx, yy , zz );

					//if ( gg->updateFlag == 0 )
					//{

					inGridParameters = getGridParameters ( inMean,inVariance,rr );

					if ( abs ( rr-inMean ) < attResolution/2 )
					{
						double obstacleP = pow ( inGridParameters.at ( 0 ),intenseFactor-attSensitivity );
						double unknownP = pow ( inGridParameters.at ( 1 ),intenseFactor-attSensitivity );
						double freeP = pow ( inGridParameters.at ( 2 ),intenseFactor-attSensitivity ) ;
						inGridParameters[0] = obstacleP;
						inGridParameters[1] = unknownP;
						inGridParameters[2] = freeP;

					}

					currentGridParameters[0] = gg->solid;
					currentGridParameters[1] = gg->unknown;
					currentGridParameters[2] = gg->free;

					newGridParameters = updateGridParameters ( inGridParameters , currentGridParameters );
					gg->solid = newGridParameters.at ( 0 );
					gg->unknown = newGridParameters.at ( 1 );
					gg->free = newGridParameters.at ( 2 );

					//gg->updateFlag=1;

					//}
				}
				rr=rr+step;
			}



			xx= ( attMapArea.xstart+attMapArea.xend ) /2;
			yy= ( attMapArea.ystart+attMapArea.yend ) /2;
			zz= ( attMapArea.zstart+attMapArea.zend ) /2;
		}


	}


	// This Time, just for setting all the accupied to zero
	xx= ( attMapArea.xstart+attMapArea.xend ) /2;
	yy= ( attMapArea.ystart+attMapArea.yend ) /2;
	zz= ( attMapArea.zstart+attMapArea.zend ) /2;

	for ( double j=-1*attHeightAngle/2 ; j<attHeightAngle/2; j=j+angleStep )
	{
		for ( double i=-1*attWidthAngle/2 ; i<attWidthAngle/2; i=i+angleStep )
		{
			for ( double rr=0.02 ; rr < attObservableDistance; rr=rr+step )
			{

// 				PosInStereo ( 0 ) =rr*cos ( j ) *cos ( i );
// 				PosInStereo ( 1 ) =rr*cos ( j ) *sin ( i );
// 				PosInStereo ( 2 ) =rr*sin ( j );

				PosInStereo ( 0 ) =rr*cos ( j ) *sin ( i );
				PosInStereo ( 1 ) =rr*sin ( j );
				PosInStereo ( 2 ) =rr*cos ( j ) *cos ( i );

				PosInStereo ( 3 ) =1;
				PosInGlobal=attSensorPosition*PosInStereo;
				xx=PosInGlobal ( 0 );
				yy=PosInGlobal ( 1 );
				zz=PosInGlobal ( 2 );
				if ( xx >attMapArea.xstart && xx<attMapArea.xend && yy>attMapArea.ystart && yy<attMapArea.yend && zz> attMapArea.zstart && zz<attMapArea.zend )
				{
					gg = xyz2grid ( xx, yy , zz );
					gg->occupied = 0;
				}
			}
			xx= ( attMapArea.xstart+attMapArea.xend ) /2;
			yy= ( attMapArea.ystart+attMapArea.yend ) /2;
			zz= ( attMapArea.zstart+attMapArea.zend ) /2;
		}
	}

#endif  //using libimage3D


//it will Free the start point of the robot
	if ( makeBoxFree ( attFreeBox.xstart,attFreeBox.ystart,attFreeBox.zstart,attFreeBox.xend,attFreeBox.yend,attFreeBox.zend ) !=WM_OK )
	{
		ODEBUG1 ( "ERROR : CwmGrid3D::updateMap - error in makeBoxFree " );
		return WM_ERROR;
	}

	return WM_OK;

}


wmStatus CwmGrid3D::allocateMap ( double xstart, double ystart, double zstart,
                                  double xend, double yend, double zend )
{
	if ( attGridMap )
		delete [] attGridMap;

	attMapArea.xstart = xstart;
	attMapArea.ystart = ystart;
	attMapArea.zstart = zstart;

	attMapArea.xend = xend;
	attMapArea.yend = yend;
	attMapArea.zend = zend;

	attMapXN = ( int ) rint ( ( xend-xstart ) /attResolution );
	attMapYN = ( int ) rint ( ( yend-ystart ) /attResolution );
	attMapZN = ( int ) rint ( ( zend-zstart ) /attResolution );

	int gridnum = attMapXN*attMapYN*attMapZN;
	attGridMap = new grid[gridnum];

	if ( clearMap() !=WM_OK )
	{
		ODEBUG1 ( "ERROR : allocateMap - error in clearMap " );
		return WM_ERROR;
	}
//initializing FreeBox! (bye defult)
	if ( setFreeBox ( xstart,ystart,zstart,xstart,ystart,zstart ) !=WM_OK )
	{
		ODEBUG1 ( "ERROR : allocateMap - error in initializing FreeBox " );
		return WM_ERROR;
	}
	return WM_OK;
}

wmStatus CwmGrid3D::setFreeBox ( double xstart, double ystart, double zstart,
                                 double xend, double yend, double zend )
{
	if ( ( xstart > xend ) || ( ystart > yend ) || ( zstart > zend ) )
	{
		ODEBUG1 ( "ERROR : setFreeBox - error in Input Data " );
		return WM_ERROR;
	}


	if ( xstart<=attMapArea.xstart )
		attFreeBox.xstart=attMapArea.xstart;
	else
		attFreeBox.xstart=xstart;

	if ( xend>=attMapArea.xend )
		attFreeBox.xend=attMapArea.xend;
	else
		attFreeBox.xend=xend;
//---------------------------------------------------
	if ( ystart<=attMapArea.ystart )
		attFreeBox.ystart=attMapArea.ystart;
	else
		attFreeBox.ystart=ystart;

	if ( yend>=attMapArea.yend )
		attFreeBox.yend=attMapArea.yend;
	else
		attFreeBox.yend=yend;
//---------------------------------------------------
	if ( zstart<=attMapArea.zstart )
		attFreeBox.zstart=attMapArea.zstart;
	else
		attFreeBox.zstart=zstart;

	if ( zend>=attMapArea.zend )
		attFreeBox.zend=attMapArea.zend;
	else
		attFreeBox.zend=zend;
	return WM_OK;

}

wmStatus CwmGrid3D::makeBoxFree ( double xstart, double ystart, double zstart,
                                  double xend, double yend, double zend )
{
	
	grid *g = attGridMap;
	for ( double zz=zstart; zz<=zend; zz=zz+attResolution/2 )
	{
		for ( double yy=ystart; yy<=yend; yy=yy+attResolution/2 )
		{
			for ( double xx=xstart; xx<=xend; xx=xx+attResolution/2 )
			{
				if ( xx >attMapArea.xstart && xx <attMapArea.xend && yy >attMapArea.ystart && yy < attMapArea.yend && zz > attMapArea.zstart && zz < attMapArea.zend )
				{

					g = xyz2grid ( xx, yy , zz );
					g->odd = 0.001;
					g->variance = 0.1;
					g->occupied = 0;
					g->check = 0;
				}
			}
		}
	}

	return WM_OK;

}

wmStatus CwmGrid3D::setRobotBoxFree ()
{
	grid *g = attGridMap;
	for ( double zz=attMapArea.zstart; zz<attMapArea.zend; zz=zz+attResolution )
	{
		for ( double yy=attMapArea.ystart; yy<attMapArea.yend; yy=yy+attResolution )
		{
			for ( double xx=attMapArea.xstart; xx<attMapArea.xend; xx=xx+attResolution )
			{
				double r = sqrt ( pow ( attRobotPosX-xx, 2 ) +pow ( attRobotPosY-yy,2 ) );
				double theta= atan2 ( yy-attRobotPosY, xx-attRobotPosX );
				if ( ( abs ( r*cos ( theta-attRobotPosTheta ) ) <= ROBOT_BOX_X/2 ) && ( abs ( r*sin ( theta-attRobotPosTheta ) ) <= ROBOT_BOX_Y/2 ) && ( zz<= ROBOT_BOX_Z ) )
				{
					g = xyz2grid ( xx, yy , zz );
					g->odd = 0.001;
					g->variance = 0.1;
					g->occupied = 0;
					g->check = 0;
				}
			}
		}
	}
	return WM_OK;
}

wmStatus CwmGrid3D::setRobotBoxPos ( double inX, double inY, double inTheta )
{
	attRobotPosX=inX;
	attRobotPosY=inY;
	attRobotPosTheta=inTheta;
	return WM_OK;
}
wmStatus CwmGrid3D::makeBoxSolid ( double xstart, double ystart, double zstart,
                                   double xend, double yend, double zend )
{

	double xS, xE,yS,yE,zS,zE;

	if ( xstart<=attMapArea.xstart )
		xS=attMapArea.xstart;
	else
		xS=xstart;

	if ( xend>=attMapArea.xend )
		xE=attMapArea.xend;
	else
		xE=xend;
//---------------------------------------------------
	if ( ystart<=attMapArea.ystart )
		yS=attMapArea.ystart;
	else
		yS=ystart;

	if ( yend>=attMapArea.yend )
		yE=attMapArea.yend;
	else
		yE=yend;
//---------------------------------------------------
	if ( zstart<=attMapArea.zstart )
		zS=attMapArea.zstart;
	else
		zS=zstart;

	if ( zend>=attMapArea.zend )
		zE=attMapArea.zend;
	else
		zE=zend;
//-----------------------------------------------------
	grid *g = attGridMap;
	for ( double zz=zS; zz<zE; zz=zz+attResolution )
	{
		for ( double yy=yS; yy<yE; yy=yy+attResolution )
		{
			for ( double xx=xS; xx<xE; xx=xx+attResolution )
			{
				g = xyz2grid ( xx, yy , zz );
				g->odd = 1000000;
				g->variance = 0.1;
				g->occupied = 0;
			}
		}
	}

	return WM_OK;

}

wmStatus CwmGrid3D::clearMap()
{
	grid *g = attGridMap;
	int gridnum = attMapXN*attMapYN*attMapZN;

	for ( int i=0; i<gridnum; i++,g++ )
	{
		//g->updateFlag = 0;
		g->odd = 1;
		g->variance = 0.1;
		g->occupied = 0;
	}
	return WM_OK;
}

wmStatus CwmGrid3D::setResolution ( double inResolution )
{
	if ( inResolution==0.0 )
		attResolution = DEFAULT_RESOLUTION;
	else
		attResolution = inResolution;

	if ( attGridMap )
	{
		allocateMap ( attMapArea.xstart, attMapArea.ystart, attMapArea.zstart,
		              attMapArea.xend, attMapArea.yend, attMapArea.zend );
	}
	else
	{
		ODEBUG1 ( "ERROR :setResolution - attGridMap does not exist " );
		return WM_ERROR;
	}

	return WM_OK;
}

wmStatus CwmGrid3D::setSensitivity ( unsigned int inSensitivity )
{
	if ( inSensitivity==0.0 )
		attSensitivity = DEFAULT_SENSITIVITY;
	else
		attSensitivity = inSensitivity;
	return WM_OK;

}
#ifdef WM3D_HAS_BOOST
wmStatus CwmGrid3D::setSensorPosition ( boost::numeric::ublas::matrix<float> inSensorPosition )
{

	attSensorPosition = inSensorPosition;
	return WM_OK;
}

wmStatus CwmGrid3D::getSensorPosition ( boost::numeric::ublas::matrix<float> outSensorPosition )
{

	outSensorPosition = attSensorPosition;
	return WM_OK;
}

wmStatus CwmGrid3D::setRangeMap ( boost::numeric::ublas::matrix<float> inRangeMap )
{

	attRangeMap=inRangeMap;
	return WM_OK;
}
#endif /* WM3D_HAS_BOOST */
#ifdef WM3D_HAS_LIBIMAGES3D
wmStatus CwmGrid3D::setSensorPosition ( const T3D &inSensorPosition )
{

	attSensorPosition = inSensorPosition;
	return WM_OK;
}

wmStatus CwmGrid3D::getSensorPosition ( T3D &outSensorPosition )
{

	outSensorPosition = attSensorPosition;
	return WM_OK;
}

wmStatus CwmGrid3D::setRangeMap ( DATA_IM3D *inRangeMap )
{

	attRangeMap=inRangeMap;
	return WM_OK;
}
#endif /* WM3D_HAS_LIBIMAGES3D */


wmStatus CwmGrid3D::refreshMap()
{
	grid *g = attGridMap;
	grid *gg = attGridMap;
// --------------------------------------------------refreh for unknown point

	int chFlag=1;
	double i=-1;
	double j=-1;
	double k=-1;
	double xx, yy, zz;

	for ( double z=attMapArea.zstart+attResolution/4;z<= ( attMapArea.zend );z=z+attResolution )
	{
		for ( double y=attMapArea.ystart+attResolution/4;y<= ( attMapArea.yend );y=y+attResolution )
		{
			for ( double x=attMapArea.xstart+attResolution/4; x <= ( attMapArea.xend );x=x+attResolution )
			{
				g=xyz2grid ( x,y,z );
				if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
				{
					k=-1;
					chFlag=1;
					while ( chFlag==1 && k<2 )
					{
						j=-1;
						while ( chFlag==1 && j<2 )
						{
							i=-1;
							while ( chFlag==1 && i<2 )
							{
								if ( ! ( i==0 && j==0 && k==0 ) )
								{
									xx=x+i*attResolution;
									yy=y+j*attResolution;
									zz=z+k*attResolution;
									if ( xx >attMapArea.xstart && xx <attMapArea.xend && yy >attMapArea.ystart && yy < attMapArea.yend && zz > attMapArea.zstart && zz < attMapArea.zend )
									{
										gg=xyz2grid ( xx,yy,zz );
										if ( ( gg->odd >= UNKNOWN_MIN ) && ( gg->odd <= UNKNOWN_MAX ) )
										{
											chFlag=0;
										}
									}
								}
								i++;
							}
							j++;
						}
						k++;
					}
					if ( chFlag==1 )
					{
						g->odd=0.001;
					}
				}
			}
		}
	}

// --------------------------------------------------refreh for solid point
	chFlag=1;
	i=-1;
	j=-1;
	k=-1;

	for ( double z=attMapArea.zstart+1*attResolution/4;z<= ( attMapArea.zend );z=z+attResolution )
	{
		for ( double y=attMapArea.ystart+attResolution/4;y<= ( attMapArea.yend );y=y+attResolution )
		{
			for ( double x=attMapArea.xstart+attResolution/4; x <= ( attMapArea.xend );x=x+attResolution )
			{
				g=xyz2grid ( x,y,z );
				if ( g->odd > UNKNOWN_MAX )
				{
					k=-1;
					chFlag=1;
					while ( chFlag==1 && k<2 )
					{
						j=-1;
						while ( chFlag==1 && j<2 )
						{
							i=-1;
							while ( chFlag==1 && i<2 )
							{
								if ( ! ( i==0 && j==0 && k==0 ) )
								{
									xx=x+i*attResolution;
									yy=y+j*attResolution;
									zz=z+k*attResolution;
									if ( xx >attMapArea.xstart && xx <attMapArea.xend && yy >attMapArea.ystart && yy < attMapArea.yend && zz > attMapArea.zstart && zz < attMapArea.zend )
									{
										gg=xyz2grid ( xx,yy,zz );
										if ( gg->odd > UNKNOWN_MAX )
										{
											chFlag=0;
										}
									}
								}
								i++;
							}
							j++;
						}
						k++;
					}
					if ( chFlag==1 )
					{
						g->odd=0.001;
					}
				}
			}
		}
	}
	return WM_OK;
}


int CwmGrid3D::getMapXN()
{
	return attMapXN;
}

int CwmGrid3D::getMapYN()
{
	return attMapYN;
}

int CwmGrid3D::getMapZN()
{
	return attMapZN;
}

grid *CwmGrid3D::getGridMap()
{
	return attGridMap;
}

wmStatus CwmGrid3D::exportOccupancyMap ( string fileName )
{

	ofstream file ( fileName.c_str() );

	file << attMapArea.xstart << " " << attMapArea.ystart << " " << attMapArea.zstart << " " << endl;
	file << attMapArea.xend << " " << attMapArea.yend << " " << attMapArea.zend << " " << endl;
	file << attMapXN << " " << attMapYN << " " << attMapZN << " " << endl;
	int NbOfGrid= attMapXN * attMapYN * attMapZN;
	file << attResolution <<" "<< NbOfGrid <<" " << 0 << " " << endl;

	for ( int i=0; i < NbOfGrid; i++ )
	{
		file << attGridMap[i].odd<<" " <<endl;
	}

	file << endl;
	file.close();

	return WM_OK;
}

wmStatus CwmGrid3D::exportVRML ( string modelFileName , string solidFileName, string unknownFileName, int exportMode )
{
	unsigned long int nbUnknown = 0;
	unsigned long int nbSolid = 0;
	cout << sizeof ( nbUnknown ) << sizeof ( 1+2*nbUnknown ) << endl;

	double R= attResolution/2;
	int NbOfGrid= attMapXN * attMapYN * attMapZN;
	double pX;
	double pY;
	double pZ;
	ofstream file ( modelFileName.c_str() );
	file << "#VRML V2.0 utf8"<< endl;
	file << "" << endl;
	file << "" << endl;
//----------------------------------------------------------------------------------------------
	for ( int i=0; i < NbOfGrid; i++ )
	{
		if ( attGridMap[i].odd>UNKNOWN_MAX )
		{
			if ( nbSolid==0 )
			{
				file <<"DEF Obstacle Shape { appearance Appearance { material Material {" << endl;
				file << "diffuseColor 1 0 0"<< endl;
				file << "specularColor 0.9 0.9 0.9" << endl;
				file << "} }" << endl;
				file << "geometry IndexedFaceSet { coordIndex["<< endl;
				file << 0+nbSolid*8 << " , " <<  1+nbSolid*8 << " , " <<  2+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 7+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  4+nbSolid*8 << " ,-1, " << 0+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  1+nbSolid*8 << " ,-1, " << 1+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  2+nbSolid*8 << " ,-1, " << 2+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 3+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  0+nbSolid*8 << " ," <<endl;
			}
			else
			{
				file <<"-1,"<< 0+nbSolid*8 << " , " <<  1+nbSolid*8 << " , " <<  2+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 7+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  4+nbSolid*8 << " ,-1, " << 0+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  1+nbSolid*8 << " ,-1, " << 1+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  2+nbSolid*8 << " ,-1, " << 2+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 3+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  0+nbSolid*8 << ","<<endl;
			}
			nbSolid ++;
		}
	}
	if ( nbSolid>0 )
	{
		file << "]"<< endl;
		file << "" << endl;
	}

	nbSolid=0;

	for ( int i=0; i < NbOfGrid; i++ )
	{
		if ( attGridMap[i].odd>UNKNOWN_MAX )
		{
			grid2xyz ( i,pX,pY,pZ );
			if ( nbSolid==0 )
			{
				file << "coord Coordinate{ point[" <<endl;
			}
			file << pX-1*R << " " << pY+R <<" " << pZ+R << ", " << pX+R << " "<< pY+R<< " " << pZ+R <<", "<< pX+R <<" "<< pY+R <<" "<< pZ-1*R << ", " << pX-1*R << " " << pY+R <<" "<< pZ-1*R <<", "<< pX-1*R << " " << pY-1*R << " " << pZ+R <<", "<<pX+R<<" "<< pY-1*R <<" "<< pZ+R <<", "<< pX+R <<" "<< pY-1*R <<" "<< pZ-1*R <<", "<< pX-1*R <<" "<< pY-1*R <<" "<< pZ-1*R <<","<< endl;
			nbSolid ++;
		}
	}
	if ( nbSolid>0 )
	{
		file << "] } } }"<< endl;
		file << "" << endl;
		file << "" << endl;
	}
//-------------------------------creating Unknown grid
	for ( int i=0; i < NbOfGrid; i++ )
	{
		if ( attGridMap[i].odd>=UNKNOWN_MIN && attGridMap[i].odd<=UNKNOWN_MAX )
		{
			if ( nbUnknown==0 )
			{
				file <<"DEF Unknown Shape { appearance Appearance { material Material {" << endl;
				file << "diffuseColor 0 1 0"<< endl;
				file << "specularColor 0.9 0.9 0.9" << endl;
				file << "} }" << endl;
				file << "geometry IndexedFaceSet { coordIndex["<< endl;
				file << 0+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 7+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " ,-1, " << 0+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " ,-1, " << 1+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " ,-1, " << 2+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 3+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  0+nbUnknown*8 << " ," <<endl;
			}
			else
			{
				file <<"-1,"<< 0+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 7+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " ,-1, " << 0+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " ,-1, " << 1+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " ,-1, " << 2+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 3+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  0+nbUnknown*8 << ","<<endl;
			}
			nbUnknown ++;
		}
	}
	if ( nbUnknown>0 )
	{
		file << "]"<< endl;
		file << "" << endl;
	}
	nbUnknown=0;

	for ( int i=0; i < NbOfGrid; i++ )
	{
		if ( attGridMap[i].odd>=UNKNOWN_MIN && attGridMap[i].odd<=UNKNOWN_MAX )
		{
			grid2xyz ( i,pX,pY,pZ );
			if ( nbUnknown==0 )
			{
				file << "coord Coordinate{ point[" <<endl;
			}
			file << pX-1*R << " " << pY+R <<" " << pZ+R << ", " << pX+R << " "<< pY+R<< " " << pZ+R <<", "<< pX+R <<" "<< pY+R <<" "<< pZ-1*R << ", " << pX-1*R << " " << pY+R <<" "<< pZ-1*R <<", "<< pX-1*R << " " << pY-1*R << " " << pZ+R <<", "<<pX+R<<" "<< pY-1*R <<" "<< pZ+R <<", "<< pX+R <<" "<< pY-1*R <<" "<< pZ-1*R <<", "<< pX-1*R <<" "<< pY-1*R <<" "<< pZ-1*R <<","<< endl;
			nbUnknown ++;
		}
	}
	if ( nbUnknown>0 )
	{
		file << "] } } }"<< endl;
		file << "" << endl;
		file << "" << endl;
	}
	file << endl;
	file.close();
	if ( exportMode == 1 )
	{
//--------------------------------create solid wrl file
		nbSolid = 0;
		ofstream solidFile ( solidFileName.c_str() );
		solidFile << "#VRML V2.0 utf8"<< endl;
		solidFile << "" << endl;
		solidFile << "" << endl;
//----------------------------------------------------------------------------------------------
		for ( int i=0; i < NbOfGrid; i++ )
		{
			if ( attGridMap[i].odd>UNKNOWN_MAX )
			{
				if ( nbSolid==0 )
				{
					solidFile <<"DEF Obstacle Shape { appearance Appearance { material Material {" << endl;
					solidFile << "diffuseColor 1 0 0"<< endl;
					solidFile << "specularColor 0.9 0.9 0.9" << endl;
					solidFile << "} }" << endl;
					solidFile << "geometry IndexedFaceSet { coordIndex["<< endl;
					solidFile << 0+nbSolid*8 << " , " <<  1+nbSolid*8 << " , " <<  2+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 7+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  4+nbSolid*8 << " ,-1, " << 0+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  1+nbSolid*8 << " ,-1, " << 1+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  2+nbSolid*8 << " ,-1, " << 2+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 3+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  0+nbSolid*8 << " ," <<endl;
				}
				else
				{
					solidFile <<"-1,"<< 0+nbSolid*8 << " , " <<  1+nbSolid*8 << " , " <<  2+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 7+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  4+nbSolid*8 << " ,-1, " << 0+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  1+nbSolid*8 << " ,-1, " << 1+nbSolid*8 << " , " <<  5+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  2+nbSolid*8 << " ,-1, " << 2+nbSolid*8 << " , " <<  6+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  3+nbSolid*8 << " ,-1, " << 3+nbSolid*8 << " , " <<  7+nbSolid*8 << " , " <<  4+nbSolid*8 << " , " <<  0+nbSolid*8 << ","<<endl;
				}
				nbSolid ++;
			}
		}
		if ( nbSolid>0 )
		{
			solidFile << "]"<< endl;
			solidFile << "" << endl;
		}
		nbSolid=0;
		for ( int i=0; i < NbOfGrid; i++ )
		{
			if ( attGridMap[i].odd>UNKNOWN_MAX )
			{
				grid2xyz ( i,pX,pY,pZ );
				if ( nbSolid==0 )
				{
					solidFile << "coord Coordinate{ point[" <<endl;
				}
				solidFile << pX-1*R << " " << pY+R <<" " << pZ+R << ", " << pX+R << " "<< pY+R<< " " << pZ+R <<", "<< pX+R <<" "<< pY+R <<" "<< pZ-1*R << ", " << pX-1*R << " " << pY+R <<" "<< pZ-1*R <<", "<< pX-1*R << " " << pY-1*R << " " << pZ+R <<", "<<pX+R<<" "<< pY-1*R <<" "<< pZ+R <<", "<< pX+R <<" "<< pY-1*R <<" "<< pZ-1*R <<", "<< pX-1*R <<" "<< pY-1*R <<" "<< pZ-1*R <<","<< endl;
				nbSolid ++;
			}
		}
		if ( nbSolid>0 )
		{
			solidFile << "] } } }"<< endl;
			solidFile << "" << endl;
			solidFile << "" << endl;
		}
		solidFile << endl;
		solidFile.close();
//------------------------------create unknown wrl file
		nbUnknown = 0;
		ofstream unknownFile ( unknownFileName.c_str() );
		unknownFile << "#VRML V2.0 utf8"<< endl;
		unknownFile << "" << endl;
		unknownFile << "" << endl;
//----------------------------------------------------------------------------------------------
		for ( int i=0; i < NbOfGrid; i++ )
		{
			if ( attGridMap[i].odd>=UNKNOWN_MIN && attGridMap[i].odd<=UNKNOWN_MAX )
			{
				if ( nbUnknown==0 )
				{
					unknownFile <<"DEF Unknown Shape { appearance Appearance { material Material {" << endl;
					unknownFile << "diffuseColor 0 1 0"<< endl;
					unknownFile << "specularColor 0.9 0.9 0.9" << endl;
					unknownFile << "} }" << endl;
					unknownFile << "geometry IndexedFaceSet { coordIndex["<< endl;
					unknownFile << 0+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 7+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " ,-1, " << 0+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " ,-1, " << 1+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " ,-1, " << 2+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 3+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  0+nbUnknown*8 << " ," <<endl;
				}
				else
				{
					unknownFile <<"-1,"<< 0+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 7+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " ,-1, " << 0+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  1+nbUnknown*8 << " ,-1, " << 1+nbUnknown*8 << " , " <<  5+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  2+nbUnknown*8 << " ,-1, " << 2+nbUnknown*8 << " , " <<  6+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  3+nbUnknown*8 << " ,-1, " << 3+nbUnknown*8 << " , " <<  7+nbUnknown*8 << " , " <<  4+nbUnknown*8 << " , " <<  0+nbUnknown*8 << ","<<endl;
				}
				nbUnknown ++;
			}
		}
		if ( nbUnknown>0 )
		{
			unknownFile << "]"<< endl;
			unknownFile << "" << endl;
		}
		nbUnknown=0;

		for ( int i=0; i < NbOfGrid; i++ )
		{
			if ( attGridMap[i].odd>=UNKNOWN_MIN && attGridMap[i].odd<=UNKNOWN_MAX )
			{
				grid2xyz ( i,pX,pY,pZ );
				if ( nbUnknown==0 )
				{
					unknownFile << "coord Coordinate{ point[" <<endl;
				}
				unknownFile << pX-1*R << " " << pY+R <<" " << pZ+R << ", " << pX+R << " "<< pY+R<< " " << pZ+R <<", "<< pX+R <<" "<< pY+R <<" "<< pZ-1*R << ", " << pX-1*R << " " << pY+R <<" "<< pZ-1*R <<", "<< pX-1*R << " " << pY-1*R << " " << pZ+R <<", "<<pX+R<<" "<< pY-1*R <<" "<< pZ+R <<", "<< pX+R <<" "<< pY-1*R <<" "<< pZ-1*R <<", "<< pX-1*R <<" "<< pY-1*R <<" "<< pZ-1*R <<","<< endl;
				nbUnknown ++;
			}
		}
		if ( nbUnknown>0 )
		{
			unknownFile << "] } } }"<< endl;
			unknownFile << "" << endl;
			unknownFile << "" << endl;
		}
		unknownFile << endl;
		unknownFile.close();
	}
//----------------report
	ODEBUG1 ( "Number of Voxel = " << NbOfGrid );
	ODEBUG1 ( "Number of Solid Voxel = " << nbSolid );
	ODEBUG1 ( "Number of Unknown Voxel = " << nbUnknown );
	ODEBUG1 ( "Number of Free Voxel = " << NbOfGrid-nbUnknown-nbSolid );
	return WM_OK;
}





wmStatus CwmGrid3D::reportCurrentStatus()
{
	cout << "------------------------------------------------------" << endl;
	cout << "           Camera Positioning Parameters                   " << endl;
	cout << "------------------------------------------------------" << endl;
	cout << "attSensorPosition =  " << endl;
	cout << attSensorPosition << endl;


	cout << "------------------------------------------------------" << endl;
	cout << "                 Grid Map  Parameters                 " << endl;
	cout << "------------------------------------------------------" << endl;

	cout << "attMapArea.Start = " << attMapArea.xstart << " , " << attMapArea.ystart << " , " << attMapArea.zstart << endl;
	cout << "attMapArea.End " << attMapArea.xend << " , " << attMapArea.yend << " , " << attMapArea.zend << endl;

	cout << "attMapXN, attMapYN, attMapZN = " << attMapXN << " , " << attMapYN << " , " << attMapZN << endl;
	cout << "attResolution = " << attResolution << endl;
	return WM_OK;
}


double CwmGrid3D::getResolution()
{
	return attResolution;
}

box_area CwmGrid3D::getMapVolume()
{
	/*	std::vector<double> mapBox;
		mapBox.resize ( 6 );
		mapBox.at ( 0 ) = attMapArea.xstart;
		mapBox.at ( 1 ) = attMapArea.ystart;
		mapBox.at ( 2 ) = attMapArea.zstart;
		mapBox.at ( 3 ) = attMapArea.xend;
		mapBox.at ( 4 ) = attMapArea.yend;
		mapBox.at ( 5 ) = attMapArea.zend;*/


	return attMapArea;
}
wmStatus CwmGrid3D::grid2xyz ( int i, double &x, double &y, double &z )
{
	i=i+1;

	int NbOfVoxelInLine= attMapXN;
	int NbOfVoxelInLayer= attMapXN*attMapYN;
	int NbOfVoxel = attMapXN * attMapYN * attMapZN;

	int lineNb, layerNb;
	if ( i <= NbOfVoxel )
	{
		lineNb=1;
		layerNb=1;
		if ( i > NbOfVoxelInLayer )
		{
			layerNb = ( i/NbOfVoxelInLayer ) +1;
			i=i%NbOfVoxelInLayer;
			if ( i==0 )
			{
				layerNb=layerNb-1;
				i=NbOfVoxelInLayer;
			}
		}

		if ( i > NbOfVoxelInLine )
		{
			lineNb = ( i/NbOfVoxelInLine ) +1;
			i=i%NbOfVoxelInLine;
			if ( i==0 )
			{
				i=NbOfVoxelInLine;
				lineNb=lineNb-1;
			}
		}

		x= attMapArea.xstart+i*attResolution-attResolution/2;
		y= attMapArea.ystart+lineNb*attResolution-attResolution/2;
		z= attMapArea.zstart+layerNb*attResolution-attResolution/2;
	}

	else
	{
		x= 0;
		y= 0;
		z= 0;
		cout << "ERROR : CwmGrid3D::grid2xyz - The grid Does Not Exist " <<endl;
		return WM_ERROR;
	}
	return WM_OK;
}


wmStatus CwmGrid3D::setCameraParameters ( double inWidthAngle , double inHeightAngle,double inObservableDistance )
{
	attWidthAngle=inWidthAngle;
	attHeightAngle=inHeightAngle;
	attObservableDistance=inObservableDistance;
	return WM_OK;
}


double CwmGrid3D::getGridParameter ( double inMean ,double inVariance, double inDistance, int k )
{
	double odd=1;
	if ( inMean > attObservableDistance+4*attResolution )
	{
		odd=0.001;
	}
	else
	{
		double discritize_step=0.01;
		if ( inVariance == 0 )
			inVariance=0.0004;
		inVariance = sqrt ( inVariance );

		double s=sqrt ( 2*M_PI );
		double probabilityOcc=0;
		for ( double d=-attResolution/2;d<=attResolution/2 ; d=d+discritize_step )
		{
			double localD=inDistance+d;
			probabilityOcc = probabilityOcc + discritize_step * ( 1/ ( inVariance*s ) * exp ( -1* ( localD-inMean ) * ( localD-inMean ) / ( 2*inVariance*inVariance ) ) );
		}
		odd=probabilityOcc/ ( 1-probabilityOcc );

		if ( k>1 )
			k=pow ( k,0.5 );
		odd = pow ( odd,k );
	}
	return odd;
}


wmStatus CwmGrid3D::importMap ( string fileName )
{
	cout << "start of CwmGrid3DTest::updateMap " << endl;
	float data[3];
	int intData[3];
	int inNbOfGrid;
	double inResolution;
	double xStart, yStart, zStart, xEnd, yEnd, zEnd;
	int in3DMapXN, in3DMapYN, in3DMapZN;
	string line,lineTemp[4];

	ifstream myfile ( fileName.c_str() );
	if ( myfile.is_open() )
	{
		cout << " File(INIT) was oppend " << endl;
		istringstream stream;
		for ( int i=0; i<4; i++ )
		{
			getline ( myfile,lineTemp[i] );
		}
		// Map Size
		stream.str ( lineTemp[0] );
		stream >> data[0] >> data[1] >> data[2];
		xStart =data[0];
		yStart =data[1];
		zStart =data[2];
		cout << "MAP SIZE :  " << endl;

		cout << "xstart = " << xStart << " , ystart = " << yStart << ", zstart = " << zStart << endl;

		stream.str ( lineTemp[1] );
		stream >> data[0] >> data[1] >> data[2];
		xEnd =data[0];
		yEnd =data[1];
		zEnd =data[2];
		cout << "xEnd = " << xEnd << " , yEnd = " << yEnd << ", zEnd = " << zEnd << endl;

		stream.str ( lineTemp[2] );
		stream >> intData[0] >> intData[1] >> intData[2];

		in3DMapXN =intData[0];
		in3DMapYN =intData[1];
		in3DMapZN =intData[2];
		cout << "in3DMapXN = " << intData[0] << " , in3DMapYN = " << intData[1] << ", in3DMapZN = " << intData[2] << endl;
		stream.str ( lineTemp[3] );
		stream >> data[0] >> intData[0]>> data[2];
		inResolution =data[0];

		inNbOfGrid=intData[0];

		cout << "Resolution = " << inResolution << endl;
		cout << "inNbOfGrid = " << inNbOfGrid << endl;

		if ( setResolution ( inResolution ) != WM_OK )
			cout <<"error in setResolution from file"<< endl;

		if ( allocateMap ( xStart,yStart,zStart,xEnd,yEnd,zEnd ) != WM_OK )
			cout <<"error in alocating 3D map from file"<< endl;

		for ( int i=0; i < inNbOfGrid; i++ )
		{
			getline ( myfile,line );
			stream.str ( line );
			stream >> data[0];
			attGridMap[i].odd=data[0];

//for importing the old map
			/*			stream >> data[0] >>data[1] >>data[2];
						if ( data[0]>data[1] && data[0]>data[2] )
						{
							attGridMap[i].odd=1;
						}
						else
						{
							if ( data[1]>data[0] && data[1]>data[2] )
								attGridMap[i].odd=1000;
							else
								attGridMap[i].odd=0.001;
						}*/
		}
	}
	else
	{
		cout << " ERROR : File(INIT) was NOT oppend " << endl;
		return WM_ERROR ;
	}
	myfile.close();

	cerr << "END - INIT." << endl;

	return WM_OK;
}

double CwmGrid3D::getVisibility ( double inX, double inY, double inTheta, double goalX, double goalY, double goalTheta )
{
# if 0 // 3D implementation
	grid *g = attGridMap;
	double angleStep=0.5*atan ( attResolution/attObservableDistance );
	double step=attResolution/2;

	for ( double theta=attWidthAngleMin+inTheta ; theta<attWidthAngleMax+inTheta; theta=theta+angleStep )
	{
		double rr=0.02;
		while ( rr < MINIMUM_ACCEPTABLE_OBSTACLE_DISTANCE )
		{
			double x= inX+rr*cos ( theta );
			double y= inY+rr*sin ( theta );
			double z= attMapArea.zstart+attResolution/10;
			if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
			{
				g = xyz2grid ( x,y,z );
				if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
				{
					ODEBUG2 ( "Config is too close to unknown  . " );
					return -1;
				}
			}
			rr=rr+step;
		}
	}

	int NbOfGrid= attMapXN * attMapYN * attMapZN;
	for ( int i=0; i < NbOfGrid; i++ )
		attGridMap[i].check=0;

	double visibility=0;
	bool obstacleFlag =0;
	for ( double j=attHeightAngleMin; j<attHeightAngleMax; j=j+angleStep )
	{
		for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
		{
			double rr=0.02;
			bool unknownVisited=false;
			obstacleFlag=0;
			while ( rr < attObservableDistance && obstacleFlag ==0 )
			{
				double x= inX+rr*cos ( j ) *cos ( i );
				double y= inY+rr*cos ( j ) *sin ( i );
				double z= CAMERA_HEIGHT+rr*sin ( j );
				if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
				{
					g = xyz2grid ( x,y,z );
					if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
					{
						if ( g->check==0 )
						{
							visibility++;
							g->check=1;
							unknownVisited=true;
						}
					}
					if ( g->odd < UNKNOWN_MIN && unknownVisited==true )
					{
						if ( g->check==0 )
						{
							visibility++;
							g->check=1;
						}
					}
					if ( ( g->odd > UNKNOWN_MAX ) )
					{
						obstacleFlag=1;
					}
				}
				rr=rr+step;
			}
		}
	}

#endif  // end of 3D implementation
# if 1 // 2D implementation
	grid *g = attGridMap;
	double angleStep=0.5*atan ( attResolution/attObservableDistance );

	double step=attResolution/2;
	ODEBUG2 ( "getVisibility : angleStep = " <<  angleStep << " , step = " << step );

//This algorithem is for preventing the robot to be too close to the unknown area
	for ( double i=0.15 ; i<0.6 ;i=i+step )
	{
		for ( double j=-0.5 ; j<0.5 ;j=j+step )
		{
			for ( double z=attMapArea.zstart ; z<attMapArea.zend ;z=z+step )
			{

				double x= inX+i*cos ( inTheta )-j*sin ( inTheta );
				double y= inY+i*sin ( inTheta ) +j*cos ( inTheta );
				if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
				{
					g = xyz2grid ( x,y,z );
					if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
					{
						ODEBUG2 ( " :getVisibility :Config is too close to unknown  . " );
						ODEBUG2 ( " :getVisibility :g->odd =  " << g->odd << " , " << UNKNOWN_MIN << " , " << UNKNOWN_MAX );
						ODEBUG2 ( " :i =" << i << " , j="<< j << " , theta = " <<  inTheta );
						ODEBUG2 ( " :X =" << x << " , Y="<< y << " , Z="<<z );
						return -1;
					}
				}
			}
		}
	}
//------------------------------------------------------------------------------------
	int nbOfGoalPixel= 0;
	int NbOfGrid= attMapXN * attMapYN * attMapZN;
	for ( int i=0; i < NbOfGrid; i++ )
		attGridMap[i].check=0;

	double visibility=0;
	bool obstacleFlag =0;
	for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
	{
		double rr=0.02;
		bool unknownVisited=false;
		obstacleFlag=0;
		bool validVoxel=true;
		while ( rr < attObservableDistance && obstacleFlag ==0 && validVoxel==true )
		{
			double x= inX+rr*cos ( i );
			double y= inY+rr*sin ( i );
			int state=0;  // Free =0, Solid=1, Unknown=2
			validVoxel=false;
			//goalVolume=0;
			for ( double z=attMapArea.zstart; z<= attMapArea.zend; z=z+attResolution )
			{
				if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
				{
					g = xyz2grid ( x,y,z );
					validVoxel=true;
					if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
						if ( state==0 )
							state=2;
					if ( g->odd > UNKNOWN_MAX )
						if ( state==0 || state==2 )
							state=1;
				}
			}

			if ( validVoxel=true )
			{
				g = xyz2grid ( x,y,attMapArea.zstart+attResolution/10 );
				if ( state==2 )
				{
					if ( g->check==0 )
					{
						visibility++;
						unknownVisited=true;
						g->check=1;
					}
				}
				if ( state==0 && unknownVisited==true )
				{
					if ( g->check==0 )
					{
						g->check=1;
						visibility++;
					}
				}
				if ( state==1 )
					obstacleFlag=1;

				if ( obstacleFlag!=1 )
				{
					if ( sqrt ( pow ( goalX-x, 2 ) +pow ( goalY-y,2 ) < .35 ) )
					{
						if ( g->check==0 )
						{
							g->check=1;
							visibility=visibility+5;
							nbOfGoalPixel++;
							ODEBUG2 ( "Config is in Goal Polygon   :" << x << " , " << y );
						}

					}
				}
			}
			rr=rr+step;
		}
	}

//for testing Goal
// 	for ( int i=0; i < NbOfGrid; i++ )
// 		attGridMap[i].check=0;
//
// 	obstacleFlag =0;
// 	bool goalVolume=0;
// 	int nbOfGoalPixel= 0;
//
// 	for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
// 	{
// 		double rr=0.02;
// 		bool unknownVisited=false;
// 		obstacleFlag=0;
// 		bool validVoxel=true;
// 		while ( rr < attObservableDistance && obstacleFlag ==0 && validVoxel==true )
// 		{
// 			double x= inX+rr*cos ( i );
// 			double y= inY+rr*sin ( i );
// 			int state=0;  // Free =0, Solid=1, Unknown=2
// 			validVoxel=false;
// 			goalVolume=0;
// 			for ( double z=attMapArea.zstart; z<= attMapArea.zend; z=z+attResolution )
// 			{
// 				if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
// 				{
// 					g = xyz2grid ( x,y,z );
// 					validVoxel=true;
// 					if ( g->odd > UNKNOWN_MAX )
// 						state=1;
// 				}
// 			}
// 			if ( validVoxel=true )
// 			{
// 				double r = sqrt ( pow ( goalX-x, 2 ) +pow ( goalY-y,2 ) );
// 				double theta= atan2 ( y-goalY, x-goalX );
// 				if ( state!=1 )
// 				{
// 					if ( ( abs ( r*cos ( theta-goalTheta ) ) <= ROBOT_X/2 ) && ( abs ( r*sin ( theta-goalTheta ) ) <= ROBOT_Y/2 ) )
// 					{
// 						if ( g->check==0 )
// 						{
// 							g->check=1;
// 							visibility=visibility+5;
// 							nbOfGoalPixel++;
// 							ODEBUG2 ( "Config is in Goal Polygon   :" << x << " , " << y );
// 						}
// 					}
// 				}
// 				else
// 				{
// 					obstacleFlag=1;
// 				}
// 			}
// 			rr=rr+step;
// 		}
// 	}
//
//


#endif  // end of 2D implementation
	ODEBUG2 ( "Visibility        is = " << visibility << " , Number of Goal Pixel =  " << nbOfGoalPixel );
	return visibility;
}

double CwmGrid3D::getVisibility ( double inX, double inY, double inTheta)
{
# if 0 // 3D implementation
	grid *g = attGridMap;
	double angleStep=0.5*atan ( attResolution/attObservableDistance );
	double step=attResolution/2;

	for ( double theta=attWidthAngleMin+inTheta ; theta<attWidthAngleMax+inTheta; theta=theta+angleStep )
	{
	double rr=0.02;
	while ( rr < MINIMUM_ACCEPTABLE_OBSTACLE_DISTANCE )
	{
	double x= inX+rr*cos ( theta );
	double y= inY+rr*sin ( theta );
	double z= attMapArea.zstart+attResolution/10;
	if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
	{
	g = xyz2grid ( x,y,z );
	if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
	{
	ODEBUG2 ( "Config is too close to unknown  . " );
				  return -1;
}
}
				  rr=rr+step;
}
}

				  int NbOfGrid= attMapXN * attMapYN * attMapZN;
				  for ( int i=0; i < NbOfGrid; i++ )
				  attGridMap[i].check=0;

				  double visibility=0;
				  bool obstacleFlag =0;
				  for ( double j=attHeightAngleMin; j<attHeightAngleMax; j=j+angleStep )
		  {
				  for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
		  {
				  double rr=0.02;
				  bool unknownVisited=false;
				  obstacleFlag=0;
				  while ( rr < attObservableDistance && obstacleFlag ==0 )
		  {
				  double x= inX+rr*cos ( j ) *cos ( i );
				  double y= inY+rr*cos ( j ) *sin ( i );
				  double z= CAMERA_HEIGHT+rr*sin ( j );
				  if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
		  {
				  g = xyz2grid ( x,y,z );
				  if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
		  {
				  if ( g->check==0 )
		  {
				  visibility++;
				  g->check=1;
				  unknownVisited=true;
}
}
				  if ( g->odd < UNKNOWN_MIN && unknownVisited==true )
		  {
				  if ( g->check==0 )
		  {
				  visibility++;
				  g->check=1;
}
}
				  if ( ( g->odd > UNKNOWN_MAX ) )
		  {
				  obstacleFlag=1;
}
}
				  rr=rr+step;
}
}
}

#endif  // end of 3D implementation
# if 1 // 2D implementation
				  grid *g = attGridMap;
		  double angleStep=0.5*atan ( attResolution/attObservableDistance );

		  double step=attResolution/2;
		  ODEBUG2 ( "getVisibility : angleStep = " <<  angleStep << " , step = " << step );

//This algorithem is for preventing the robot to be too close to the unknown area
		  for ( double i=0.15 ; i<0.6 ;i=i+step )
		  {
			  for ( double j=-0.5 ; j<0.5 ;j=j+step )
			  {
				  for ( double z=attMapArea.zstart ; z<attMapArea.zend ;z=z+step )
				  {

					  double x= inX+i*cos ( inTheta )-j*sin ( inTheta );
					  double y= inY+i*sin ( inTheta ) +j*cos ( inTheta );
					  if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
					  {
						  g = xyz2grid ( x,y,z );
						  if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
						  {
							  ODEBUG2 ( " :getVisibility :Config is too close to unknown  . " );
							  ODEBUG2 ( " :getVisibility :g->odd =  " << g->odd << " , " << UNKNOWN_MIN << " , " << UNKNOWN_MAX );
							  ODEBUG2 ( " :i =" << i << " , j="<< j << " , theta = " <<  inTheta );
							  ODEBUG2 ( " :X =" << x << " , Y="<< y << " , Z="<<z );
							  return -1;
						  }
					  }
				  }
			  }
		  }
//------------------------------------------------------------------------------------
		  int nbOfGoalPixel= 0;
		  int NbOfGrid= attMapXN * attMapYN * attMapZN;
		  for ( int i=0; i < NbOfGrid; i++ )
			  attGridMap[i].check=0;

		  double visibility=0;
		  bool obstacleFlag =0;
		  for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
		  {
			  double rr=0.02;
			  bool unknownVisited=false;
			  obstacleFlag=0;
			  bool validVoxel=true;
			  while ( rr < attObservableDistance && obstacleFlag ==0 && validVoxel==true )
			  {
				  double x= inX+rr*cos ( i );
				  double y= inY+rr*sin ( i );
				  int state=0;  // Free =0, Solid=1, Unknown=2
				  validVoxel=false;
			//goalVolume=0;
				  for ( double z=attMapArea.zstart; z<= attMapArea.zend; z=z+attResolution )
				  {
					  if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
					  {
						  g = xyz2grid ( x,y,z );
						  validVoxel=true;
						  if ( ( g->odd >= UNKNOWN_MIN ) && ( g->odd <= UNKNOWN_MAX ) )
							  if ( state==0 )
								  state=2;
						  if ( g->odd > UNKNOWN_MAX )
							  if ( state==0 || state==2 )
								  state=1;
					  }
				  }

				  if ( validVoxel=true )
				  {
					  g = xyz2grid ( x,y,attMapArea.zstart+attResolution/10 );
					  if ( state==2 )
					  {
						  if ( g->check==0 )
						  {
							  visibility++;
							  unknownVisited=true;
							  g->check=1;
						  }
					  }
					  if ( state==0 && unknownVisited==true )
					  {
						  if ( g->check==0 )
						  {
							  g->check=1;
							  visibility++;
						  }
					  }
					  if ( state==1 )
						  obstacleFlag=1;

				  }
				  rr=rr+step;
			  }
		  }

//for testing Goal
// 	for ( int i=0; i < NbOfGrid; i++ )
// 		attGridMap[i].check=0;
		  //
// 	obstacleFlag =0;
// 	bool goalVolume=0;
// 	int nbOfGoalPixel= 0;
		  //
// 	for ( double i= attWidthAngleMin+inTheta ; i< attWidthAngleMax+inTheta; i=i+angleStep )
// 	{
// 		double rr=0.02;
// 		bool unknownVisited=false;
// 		obstacleFlag=0;
// 		bool validVoxel=true;
// 		while ( rr < attObservableDistance && obstacleFlag ==0 && validVoxel==true )
// 		{
// 			double x= inX+rr*cos ( i );
// 			double y= inY+rr*sin ( i );
// 			int state=0;  // Free =0, Solid=1, Unknown=2
// 			validVoxel=false;
// 			goalVolume=0;
// 			for ( double z=attMapArea.zstart; z<= attMapArea.zend; z=z+attResolution )
// 			{
// 				if ( x >attMapArea.xstart && x<attMapArea.xend && y>attMapArea.ystart && y<attMapArea.yend && z> attMapArea.zstart && z<attMapArea.zend )
// 				{
// 					g = xyz2grid ( x,y,z );
// 					validVoxel=true;
// 					if ( g->odd > UNKNOWN_MAX )
// 						state=1;
// 				}
// 			}
// 			if ( validVoxel=true )
// 			{
// 				double r = sqrt ( pow ( goalX-x, 2 ) +pow ( goalY-y,2 ) );
// 				double theta= atan2 ( y-goalY, x-goalX );
// 				if ( state!=1 )
// 				{
// 					if ( ( abs ( r*cos ( theta-goalTheta ) ) <= ROBOT_X/2 ) && ( abs ( r*sin ( theta-goalTheta ) ) <= ROBOT_Y/2 ) )
// 					{
// 						if ( g->check==0 )
// 						{
// 							g->check=1;
// 							visibility=visibility+5;
// 							nbOfGoalPixel++;
// 							ODEBUG2 ( "Config is in Goal Polygon   :" << x << " , " << y );
// 						}
// 					}
// 				}
// 				else
// 				{
// 					obstacleFlag=1;
// 				}
// 			}
// 			rr=rr+step;
// 		}
// 	}
		  //
		  //


#endif  // end of 2D implementation
		  ODEBUG2 ( "Visibility        is = " << visibility << " , Number of Goal Pixel =  " << nbOfGoalPixel );
		  return visibility;
}


wmStatus CwmGrid3D::setCoveringAngle ( double inHeightAngleMin, double inHeightAngleMax, double inWidthAngleMin, double inWidthAngleMax )
{
	attHeightAngleMin=inHeightAngleMin;
	attHeightAngleMax=inHeightAngleMax;
	attWidthAngleMin=inWidthAngleMin;
	attWidthAngleMax=inWidthAngleMax;
	return WM_OK;
}

double CwmGrid3D::getObservableDistance()
{
	return attObservableDistance;
}

wmStatus CwmGrid3D::setGridState ( std::vector<char> inStateVector )
{
	for ( unsigned int i=0; i < inStateVector.size(); i++ )
	{
		switch ( inStateVector[i] )
		{
			case 0:
			{
				attGridMap[i].odd=0.001;
				break;
			}
			case 1:
			{
				attGridMap[i].odd=1000;
				break;
			}
			case 2:
			{
				attGridMap[i].odd=1;
				break;
			}
		}
	}
	return WM_OK;
}
